//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

#include "Particle.hh"

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

extern vector <Wall2D *> walls;
extern bool draw_textures;
extern bool draw_intersections;

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

// initialize particle's position, velocity, mass, radius

Particle2D::Particle2D(double x0, double y0, double vx0, double vy0, double mass, double radius, double rest, 
		       bool wall_collide,
		       int texture_num, float re, float gr, float bl)
{ 
  set(x0, y0, vx0, vy0);

  m = mass;
  r = radius;
  restitution = rest;

  collides_with_walls = wall_collide;

  red = re;
  green = gr;
  blue = bl;

  hit_wall_flag = false;

  // we don't use the first two numbers here for rendering--only the radius r matters

  sprite = new Sprite(16, 16, 1, 1, 1, texture_num);
}

//----------------------------------------------------------------------------

// directly set particle's position, velocity

void Particle2D::set(double x0, double y0, double vx0, double vy0)
{ 
  p.x = old_p.x = x0;
  p.y = old_p.y = y0;

  v.x = old_v.x = vx0;
  v.y = old_v.y = vy0;

  a.x = a.y = 0.0;
}

//----------------------------------------------------------------------------

// returns true if there was a collision, false otherwise.
// particle_t is how far along particle path from old_p to p collision occurs,
// if it does at all.  0 means right at old_p, 1 means p.

bool Particle2D::detect_wall_collision(Wall2D *wall, double & particle_t)
{ 
  Point2D particle_delta;
  Point2D reflect;
  double particle_num, wall_num, denom;
  double wall_t;   // how far along wall segment collision occurs

  // intersect particle path line segment with wall line segment.
  // these approach is taken from http://local.wasp.uwa.edu.au/~pbourke/geometry/lineline2d/

  particle_delta.x = p.x - old_p.x;
  particle_delta.y = p.y - old_p.y;

  denom = wall->delta.y * particle_delta.x - wall->delta.x * particle_delta.y;

  // lines are parallel

  if (!denom)
    return false;

  particle_num = wall->delta.x * (old_p.y - wall->p1.y) - wall->delta.y * (old_p.x - wall->p1.x);
  wall_num = particle_delta.x * (old_p.y - wall->p1.y) - particle_delta.y * (old_p.x - wall->p1.x);

  particle_t = particle_num / denom;
  wall_t = wall_num / denom;

  // intersection is NOT within at least one of the segments

  if (particle_t < 0 || particle_t > 1.0 || wall_t < 0 || wall_t > 1.0)
    return false;

  // intersection IS within both segments

  else
    return true;
}

//----------------------------------------------------------------------------

// particle_t is how far along particle path from old_p to p collision occurs.
// 0 means right at old_p, 1 means p.

void Particle2D::handle_wall_collision(Wall2D *wall, double particle_t) 
{ 
  Point2D reflect;
  double dotprod, velmag, refmag;

  // compute intersection point

  inter.x = old_p.x + particle_t * (p.x - old_p.x);
  inter.y = old_p.y + particle_t * (p.y - old_p.y);

  // reflect the portion of the path that occurred after the collision

  reflect.x = p.x - inter.x;
  reflect.y = p.y - inter.y;

  dotprod = wall->normal.x * reflect.x + wall->normal.y * reflect.y;

  reflect.x -= 2.0 * dotprod * wall->normal.x;
  reflect.y -= 2.0 * dotprod * wall->normal.y;

  velmag = sqrt(old_v.x * old_v.x + old_v.y * old_v.y);

  refmag = sqrt(reflect.x * reflect.x + reflect.y * reflect.y);

  // new start point, end point, and velocity

  p.x = inter.x + reflect.x;
  p.y = inter.y + reflect.y;

  old_p.x = inter.x;
  old_p.y = inter.y;

  // scale speed relative to dt and apply a little inelasticity.
  // really there should be different values in the normal vs. tangent
  // directions, but....

  old_v.x = restitution * reflect.x * velmag / refmag;
  old_v.y = restitution * reflect.y * velmag / refmag;  
}

//----------------------------------------------------------------------------

// this is the method where we back up to the exact spot of the collision and apply 
// the vector reflection formula in handle_collision()--it's NOT the penalty method

// may change p, old_p, and old_v

void Particle2D::compute_wall_collisions()
{
  double particle_t, closest_particle_t;
  int i, closest_wall;

  // while loop here because we may hit multiple walls before reaching new position

  while (1) {

    closest_particle_t = 1.1;

    // did particle hit any wall?
    
    for (i = 0; i < walls.size(); i++) 
      if (detect_wall_collision(walls[i], particle_t)) {
	if (particle_t > 0.0000001 && particle_t < closest_particle_t) { // the 0.000whatever1 here is to deal with a precision issue; otherwise we 
	                                                                 // may think we're hitting a new wall when it's really the one we just hit
	  closest_particle_t = particle_t;
	  closest_wall = i;
	}
      }
    
    // yes
    
    if (closest_particle_t <= 1.0) {
      hit_wall_flag = true;   // we don't reset this to false so we always have the last hit location 
      handle_wall_collision(walls[closest_wall], closest_particle_t);
    }
    
    // no

    else
      break;
  }
}

//----------------------------------------------------------------------------

// add acceleration due to one force to accumulator

void Particle2D::add_acceleration(double ax, double ay)
{
  a.x += ax;
  a.y += ay;
}

//----------------------------------------------------------------------------

// subject to net acceleration (ax, ay), move particle to new position after dt seconds have elapsed

void Particle2D::update(double dt)
{
  // remember old values

  old_p.x = p.x;
  old_p.y = p.y;

  old_v.x = v.x;
  old_v.y = v.y;

  // what would new position be if there was no collision along the way? -- Euler method

  p.x = old_p.x + dt * old_v.x;
  p.y = old_p.y + dt * old_v.y;

  // now see if we actually did hit anything, and if so deal with it 
  // (may change p, old_p, and old_v)

  if (collides_with_walls)
    compute_wall_collisions();

  // what is new velocity? -- Euler method

  v.x = old_v.x + dt * a.x;
  v.y = old_v.y + dt * a.y;

  // clear accelerations

  a.x = a.y = 0.0;
}

//----------------------------------------------------------------------------

// render particle

void Particle2D::draw()
{
  if (draw_textures) {
    glColor3f(red, green, blue);
    sprite->draw(p.x, p.y, 0, r, r);
  }
  else {
    glPointSize(5);

    glBegin(GL_POINTS);
    
    glColor3f(1, 1, 1);
    glVertex2f(p.x, p.y);

    // draw point of last wall collision
    
    if (draw_intersections && hit_wall_flag) {
      glColor3f(0, 1, 0);
      glVertex2f(inter.x, inter.y);
    }
    
    glEnd();
  }
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
