//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//
// "Bounce!" v. 1.1
//
// by Christopher Rasmussen
//
// CISC 440/640, 2012 
//
// A basic 2-D particle system with gravity, drag, and planar collisions
//
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>

#include "Sprite.hh"
#include "Wall.hh"
#include "Particle.hh"

#include <math.h>
#include <string.h>
#include <sys/time.h>

#include <GL/glut.h>

#include <vector>

using namespace std;

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

void initialize_random();
double uniform_random(double, double);
int uniform_random_int(int, int);
void mouse_to_meters(int, int, double &, double &);
void set_blend();

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

#define TEXTURE_BLUEBALL_16x16     1
#define TEXTURE_FIREBALL           2
#define TEXTURE_PARTICLE_32x32     3

int win_w = 400;
int win_h = 400;

double x_min = 0.0;
double x_max = 10.0;
double y_min = 0.0;
double y_max = 10.0;

double timestep = 0.01;   // seconds

bool paused = false;
bool draw_textures = true;
bool draw_intersections = false;

vector <Wall2D *> walls;
bool new_wall_flag = false;
Point2D new_wall;

int num_balls = 100;
vector <Particle2D *> balls;

bool do_blend = false;
bool do_trails = false;

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

// set seed based on time so no two games are alike

void initialize_random()
{
  struct timeval tp;

  gettimeofday(&tp, NULL);
  srand48(tp.tv_sec);
}

//----------------------------------------------------------------------------

// random float in range [lower, upper]

double uniform_random(double lower, double upper)
{
  double result;
  double range_size;

  range_size = upper - lower;
  result = range_size * drand48();
  result += lower;

  return result;
}

//----------------------------------------------------------------------------

// random int in range [lower, upper]

int uniform_random_int(int lower, int upper)
{
  return (int) round(uniform_random(lower, upper));
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

void mouse_to_meters(int x, int y, double & mx, double & my)
{
  mx = (x_max - x_min) * (double) x / (double) win_w;
  my = (y_max - y_min) * (double) (win_h - y) / (double) win_h;
}

//----------------------------------------------------------------------------

void initialize_bounce()
{
  int i;

  initialize_random();

  // textures

  int numTextures = 3;
  char **texfilename = (char **) malloc(numTextures * sizeof(char *));
  for (i = 0; i < numTextures; i++) 
    texfilename[i] = (char *) malloc(256 * sizeof(char));

  sprintf(texfilename[0], "blueball_16x16.ppm");
  sprintf(texfilename[1], "fireball.ppm");
  sprintf(texfilename[2], "particletexture_32x32.ppm");

  load_sprites(numTextures, texfilename);

  // initialize walls

  walls.clear();

  // floor

  walls.push_back(new Wall2D(x_min + 0.5, y_min + 0.5,
			     x_max - 0.5, y_min + 0.5));

  // sides

  walls.push_back(new Wall2D(x_max - 0.5, y_min + 0.5,
			     x_max - 0.5, y_max - 0.5));

  walls.push_back(new Wall2D(x_min + 0.5, y_min + 0.5,
			     x_min + 0.5, y_max - 0.5));
	  
  // ceiling

  walls.push_back(new Wall2D(x_min + 0.5, y_max - 0.5,
			     x_max - 0.5, y_max - 0.5));

  // initialize balls

  balls.clear();

  for (i = 0; i < num_balls; i++)
    balls.push_back(new Particle2D(0.5 * (x_min + x_max), 8.0, 
 				   uniform_random(-2.0, 2.0), uniform_random(0.0, 3.0),
				   //    balls.push_back(new Particle2D(0.5 * (x_min + x_max), 1.0, 
				   //				   uniform_random(-1.0, 2.5), uniform_random(10.0, 12.0),
				   DEFAULT_MASS, 
				   .1,
				   //				   uniform_random(.1, 1.0), 
				   //DEFAULT_RADIUS, 
				   DEFAULT_RESTITUTION,
				   true,
				   TEXTURE_BLUEBALL_16x16));
  //TEXTURE_PARTICLE_32x32));
  //TEXTURE_FIREBALL));
  //  				   uniform_random(0.0, 1.0), uniform_random(0.0, 1.0), uniform_random(0.0, 1.0)));
}

//----------------------------------------------------------------------------

// has a key been pressed? 

void keyboard(unsigned char key, int x, int y)
{
  int i;

  // show collision locations ("intersections")--only works when texturing off

  if (key == 'i') 
    draw_intersections = !draw_intersections;

  // clear walls (all but container)

  else if (key == 'c') {
    for (i = walls.size() - 1; i >= 4; i--)
      walls.erase(walls.begin() + i);

    //    walls.clear();   // memory leak?
  }

  // toggle textures

  else if (key == 't')
    draw_textures = !draw_textures;

  // toggle blending

  else if (key == 'b') {
    do_blend = !do_blend;
    set_blend();
  }

  // pause

  else if (key == 'p')
    paused = !paused;

  // quit

  else if (key == 'q')
    exit(1);

  // reset ball locations

  else if (key == 'r') {
    for (i = 0; i < balls.size(); i++) {
      balls[i]->set(0.5 * (x_min + x_max), 8.0, 
		    uniform_random(-2.0, 2.0), uniform_random(0.0, 3.0));
    }
  }
}

//----------------------------------------------------------------------------

// no other events

void idle(int value)
{
  int i;

  if (!paused) {

    for (i = 0; i < balls.size(); i++) {
      balls[i]->add_acceleration(0.0, -9.8);                                    // gravity 
      balls[i]->add_acceleration(-0.1 * balls[i]->v.x, -0.1 * balls[i]->v.y);   // drag
      balls[i]->update(timestep);   
    }
  }

  glutPostRedisplay();

  glutTimerFunc((int) (1000.0 * timestep), idle, 0);
}

//----------------------------------------------------------------------------

// use mouse button to place walls

void mouse(int button, int state, int x, int y)
{
  double mx, my;

  if (state == GLUT_UP) {

    mouse_to_meters(x, y, mx, my);
    
    if (button == GLUT_LEFT_BUTTON) {

      if (!new_wall_flag) {
	new_wall.x = mx;
	new_wall.y = my;
	new_wall_flag = true;
      }
      else {
	walls.push_back(new Wall2D(new_wall.x, new_wall.y, mx, my));
	new_wall.x = mx;
	new_wall.y = my;
      }
    }
    else if (button == GLUT_RIGHT_BUTTON) {
      if (new_wall_flag) {
	walls.push_back(new Wall2D(new_wall.x, new_wall.y, mx, my));
	new_wall_flag = false;
      }
    }

    glutPostRedisplay();
  }
}

//----------------------------------------------------------------------------

void display() 
{
  int i;

  // clear background

  glClear(GL_COLOR_BUFFER_BIT);

  if (do_trails) {
    glAccum(GL_RETURN, 0.9);
    glClear(GL_ACCUM_BUFFER_BIT);
  }

  // draw

  if (new_wall_flag) {

    glPointSize(5);
    glColor3f(0, 1, 0);

    glBegin(GL_POINTS);

    glVertex2f(new_wall.x, new_wall.y);

    glEnd();

  }

  for (i = 0; i < walls.size(); i++)
    walls[i]->draw();

  if (draw_textures) {
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
  }

  for (i = 0; i < balls.size(); i++)
    balls[i]->draw();

  if (draw_textures) {
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);
  }

  // finish up

  glutSwapBuffers();

  if (do_trails)
    glAccum(GL_ACCUM, 0.99);

}

//----------------------------------------------------------------------------

// this is helpful for telling the different blend modes apart: http://www.andersriggelsen.dk/glblendfunc.php

void set_blend()
{
  // blended together

  if (do_blend) 
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);

  // individual sprites

  else
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

//----------------------------------------------------------------------------

void init(int argc, char** argv) {

  glClearColor(0.0, 0.0, 0.0, 0.0);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluOrtho2D(x_min, x_max, y_min, y_max);

  for (int i = 0; i < argc; i++) {
    if (!strcmp("-blend", argv[i]))
      do_blend = true;
    else if (!strcmp("-trails", argv[i]))
      do_trails = true;
   else if (!strcmp("-n", argv[i]) && argc > i + 1)
     num_balls = atoi(argv[i + i]);
  }

  set_blend();

  if (do_trails)
    glClear(GL_ACCUM_BUFFER_BIT);
}

//----------------------------------------------------------------------------

int main(int argc, char** argv)
{
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
  glutInitWindowSize(win_w, win_h);
  glutCreateWindow("bounce");

  // set OpenGL states, variables

  init(argc, argv); 			            

  // set program-specific variables

  initialize_bounce();

  // register callback routines

  glutDisplayFunc(display);
  glutKeyboardFunc(keyboard);
  glutMouseFunc(mouse);
  glutTimerFunc((int) (1000.0 * timestep), idle, 0);

  // enter event-driven loop

  glutMainLoop(); 		    

  return 1;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
